import pandas as pd
import pytest

import skrub._dataframe._pandas as skrub_pd
import skrub._dataframe._polars as skrub_pl
from skrub._dataframe._namespace import get_df_namespace
from skrub.conftest import _POLARS_INSTALLED

main = pd.DataFrame(
    {
        "userId": [1, 1, 1, 2, 2, 2],
        "movieId": [1, 3, 6, 318, 6, 1704],
        "rating": [4.0, 4.0, 4.0, 3.0, 2.0, 4.0],
        "genre": ["drama", "drama", "comedy", "sf", "comedy", "sf"],
    }
)


def test_get_namespace_pandas():
    skrub_px, px = get_df_namespace(main, main)
    assert skrub_px is skrub_pd
    assert px is pd

    with pytest.raises(TypeError, match=r"(?=.*Only Pandas or Polars)(?=.*supported)"):
        get_df_namespace(main, main.values)


@pytest.mark.skipif(not _POLARS_INSTALLED, reason="Polars is not available")
def test_get_namespace_polars():
    import polars as pl

    skrub_px, px = get_df_namespace(pl.DataFrame(main), pl.DataFrame(main))
    assert skrub_px is skrub_pl
    assert px is pl

    with pytest.raises(TypeError, match=r"(?=.*Mixing Pandas)(?=.*Polars)"):
        get_df_namespace(main, pl.DataFrame(main))

    with pytest.raises(
        TypeError, match=r"(?=.*Mixing)(?=.*lazyframes)(?=.*dataframes)"
    ):
        get_df_namespace(pl.DataFrame(main), pl.LazyFrame(main))



from datetime import datetime, timezone

import pandas as pd
from sklearn.utils.validation import check_is_fitted

try:
    import polars as pl
except ImportError:
    pass

from . import _dataframe as sbd
from ._dispatch import dispatch
from ._on_each_column import RejectColumn, SingleColumnTransformer

__all__ = ["DatetimeEncoder"]

_TIME_LEVELS = [
    "year",
    "month",
    "day",
    "hour",
    "minute",
    "second",
    "microsecond",
    "nanosecond",
]


@dispatch
def _is_date(col):
    raise NotImplementedError()


@_is_date.specialize("pandas", argument_type="Column")
def _is_date_pandas(col):
    col = sbd.drop_nulls(col)
    return (col.dt.date == col).all()


@_is_date.specialize("polars", argument_type="Column")
def _is_date_polars(col):
    return (col.dt.date() == col).all()


@dispatch
def _get_dt_feature(col, feature):
    raise NotImplementedError()


@_get_dt_feature.specialize("pandas", argument_type="Column")
def _get_dt_feature_pandas(col, feature):
    if feature == "total_seconds":
        if col.dt.tz is None:
            epoch = datetime(1970, 1, 1)
        else:
            epoch = datetime(1970, 1, 1, tzinfo=timezone.utc)
        return ((col - epoch) / pd.Timedelta("1s")).astype("float32")
    if feature == "weekday":
        return col.dt.day_of_week + 1
    assert feature in _TIME_LEVELS
    return getattr(col.dt, feature)


@_get_dt_feature.specialize("polars", argument_type="Column")
def _get_dt_feature_polars(col, feature):
    if feature == "total_seconds":
        return (col.dt.timestamp(time_unit="ms") / 1000).cast(pl.Float32)
    assert feature in _TIME_LEVELS + ["weekday"]
    return getattr(col.dt, feature)()


class DatetimeEncoder(SingleColumnTransformer):
    """
    Extract temporal features such as month, day of the week, … from a datetime column.

    All extracted features are provided as float32 columns.

    No timezone conversion is performed: if the input column is timezone aware, the
    extracted features will be in the column's timezone.

    An input column that does not have a Date or Datetime dtype will be
    rejected by raising a ``RejectColumn`` exception. See ``ToDatetime`` for
    converting strings to proper datetimes. **Note:** the ``TableVectorizer``
    only sends datetime columns to its ``datetime_encoder``. Therefore it is
    always safe to use a ``DatetimeEncoder`` as the ``TableVectorizer``'s
    ``datetime_encoder`` parameter.

    Parameters
    ----------
    resolution : str or None, default="hour"
        If a string, extract up to this resolution. Must be "year", "month",
        "day", "hour", "minute", "second", "microsecond", or "nanosecond". For
        example, ``resolution="day"`` generates the features "year", "month",
        and "day" only. If the input column contains dates with no time
        information, time features ("hour", "minute", … ) are never extracted.
        If ``None``, the features listed above are not extracted (but day of
        the week and total seconds may still be extracted, see below).

    add_weekday : bool, default=False
        Extract the day of the week as a numerical feature from 1 (Monday) to 7
        (Sunday).

    add_total_seconds : bool, default=True
        Add the total number of seconds since the Unix epoch (00:00:00 UTC on 1
        January 1970).

    Attributes
    ----------
    extracted_features_ : list of strings
        The features that are extracted, a subset of ["year", …, "nanosecond",
        "weekday", "total_seconds"]

    See Also
    --------
    ToDatetime :
        Convert strings to datetimes.

    Examples
    --------
    >>> import pandas as pd

    >>> login = pd.to_datetime(
    ...     pd.Series(
    ...         ["2024-05-13T12:05:36", None, "2024-05-15T13:46:02"], name="login")
    ... )
    >>> login
    0   2024-05-13 12:05:36
    1                   NaT
    2   2024-05-15 13:46:02
    Name: login, dtype: datetime64[ns]
    >>> from skrub import DatetimeEncoder

    >>> DatetimeEncoder().fit_transform(login)
       login_year  login_month  login_day  login_hour  login_total_seconds
    0      2024.0          5.0       13.0        12.0         1.715602e+09
    1         NaN          NaN        NaN         NaN                  NaN
    2      2024.0          5.0       15.0        13.0         1.715781e+09

    We can ask for a finer resolution:

    >>> DatetimeEncoder(resolution='second', add_total_seconds=False).fit_transform(
    ...     login
    ... )
       login_year  login_month  login_day  login_hour  login_minute  login_second
    0      2024.0          5.0       13.0        12.0           5.0          36.0
    1         NaN          NaN        NaN         NaN           NaN           NaN
    2      2024.0          5.0       15.0        13.0          46.0           2.0

    We can also ask for the day of the week. The week starts at 1 on Monday and ends
    at 7 on Sunday. This is consistent with the ISO week date system
    (https://en.wikipedia.org/wiki/ISO_week_date), the standard library
    ``datetime.isoweekday()`` and polars ``weekday``, but not with pandas
    ``day_of_week``, which counts days from 0.

    >>> login.dt.strftime('%A = %w')
    0       Monday = 1
    1              NaN
    2    Wednesday = 3
    Name: login, dtype: object
    >>> login.dt.day_of_week
    0    0.0
    1    NaN
    2    2.0
    Name: login, dtype: float64
    >>> DatetimeEncoder(add_weekday=True, add_total_seconds=False).fit_transform(login)
       login_year  login_month  login_day  login_hour  login_weekday
    0      2024.0          5.0       13.0        12.0            1.0
    1         NaN          NaN        NaN         NaN            NaN
    2      2024.0          5.0       15.0        13.0            3.0

    When a column contains only dates without time information, the time features
    are discarded, regardless of ``resolution``.

    >>> birthday = pd.to_datetime(
    ...     pd.Series(['2024-04-14', '2024-05-15'], name='birthday')
    ... )
    >>> encoder = DatetimeEncoder(resolution='second')
    >>> encoder.fit_transform(birthday)
       birthday_year  birthday_month  birthday_day  birthday_total_seconds
    0         2024.0             4.0          14.0            1.713053e+09
    1         2024.0             5.0          15.0            1.715731e+09
    >>> encoder.extracted_features_
    ['year', 'month', 'day', 'total_seconds']

    (The number of seconds since Epoch can still be extracted but not "hour",
    "minute", etc.)

    Non-datetime columns are rejected by raising a ``RejectColumn`` exception.

    >>> s = pd.Series(['2024-04-14', '2024-05-15'], name='birthday')
    >>> s
    0    2024-04-14
    1    2024-05-15
    Name: birthday, dtype: object
    >>> DatetimeEncoder().fit_transform(s)
    Traceback (most recent call last):
        ...
    skrub._on_each_column.RejectColumn: Column 'birthday' does not have Date or Datetime dtype.

    :class:`ToDatetime`: can be used for converting strings to datetimes.

    >>> from skrub import ToDatetime
    >>> from sklearn.pipeline import make_pipeline
    >>> make_pipeline(ToDatetime(), DatetimeEncoder()).fit_transform(s)
       birthday_year  birthday_month  birthday_day  birthday_total_seconds
    0         2024.0             4.0          14.0            1.713053e+09
    1         2024.0             5.0          15.0            1.715731e+09

    **Time zones**

    If the input column has a time zone, the extracted features are in this timezone.

    >>> login = pd.to_datetime(
    ...     pd.Series(
    ...         ["2024-05-13T12:05:36", None, "2024-05-15T13:46:02"], name="login")
    ... ).dt.tz_localize('Europe/Paris')
    >>> encoder = DatetimeEncoder()
    >>> encoder.fit_transform(login)['login_hour']
    0    12.0
    1     NaN
    2    13.0
    Name: login_hour, dtype: float32

    No special care is taken to convert inputs to ``transform`` to the same time
    zone as the column the encoder was fitted on. The features are always in the
    time zone of the input.

    >>> login_sp = login.dt.tz_convert('America/Sao_Paulo')
    >>> login_sp
    0   2024-05-13 07:05:36-03:00
    1                         NaT
    2   2024-05-15 08:46:02-03:00
    Name: login, dtype: datetime64[ns, America/Sao_Paulo]
    >>> encoder.transform(login_sp)['login_hour']
    0    7.0
    1    NaN
    2    8.0
    Name: login_hour, dtype: float32

    To ensure datetime columns are in a consistent timezones, use ``ToDatetime``.

    >>> encoder = make_pipeline(ToDatetime(), DatetimeEncoder())
    >>> encoder.fit_transform(login)['login_hour']
    0    12.0
    1     NaN
    2    13.0
    Name: login_hour, dtype: float32
    >>> encoder.transform(login_sp)['login_hour']
    0    12.0
    1     NaN
    2    13.0
    Name: login_hour, dtype: float32

    Here we can see the input to ``transform`` has been converted back to the
    timezone used during ``fit`` and that we get the same result for "hour".
    """  # noqa: E501

    def __init__(self, resolution="hour", add_weekday=False, add_total_seconds=True):
        self.resolution = resolution
        self.add_weekday = add_weekday
        self.add_total_seconds = add_total_seconds

    def fit_transform(self, column, y=None):
        """Fit the encoder and transform a column.

        Parameters
        ----------
        column : pandas or polars Series with dtype Date or Datetime
            The input to transform.

        y : None
            Ignored.

        Returns
        -------
        transformed : DataFrame
            The extracted features.
        """
        del y
        self._check_params()
        if not sbd.is_any_date(column):
            raise RejectColumn(
                f"Column {sbd.name(column)!r} does not have Date or Datetime dtype."
            )
        if self.resolution is None:
            self.extracted_features_ = []
        else:
            idx_level = _TIME_LEVELS.index(self.resolution)
            if _is_date(column):
                idx_level = min(idx_level, _TIME_LEVELS.index("day"))
            self.extracted_features_ = _TIME_LEVELS[: idx_level + 1]
        if self.add_total_seconds:
            self.extracted_features_.append("total_seconds")
        if self.add_weekday:
            self.extracted_features_.append("weekday")
        return self.transform(column)

    def transform(self, column):
        """Transform a column.

        Parameters
        ----------
        column : pandas or polars Series with dtype Date or Datetime
            The input to transform.

        Returns
        -------
        transformed : DataFrame
            The extracted features.
        """
        check_is_fitted(self, "extracted_features_")
        name = sbd.name(column)
        all_extracted = []
        for feature in self.extracted_features_:
            extracted = _get_dt_feature(column, feature).rename(f"{name}_{feature}")
            extracted = sbd.to_float32(extracted)
            all_extracted.append(extracted)
        return sbd.make_dataframe_like(column, all_extracted)

    def _check_params(self):
        allowed = _TIME_LEVELS + [None]
        if self.resolution not in allowed:
            raise ValueError(
                f"'resolution' options are {allowed}, got {self.resolution!r}."
            )



from pathlib import Path


def get_data_home(data_home=None):
    """Returns the path of the skrub data directory.

    This folder is used by some large dataset loaders to avoid downloading the
    data several times.

    By default the data directory is set to a folder named 'skrub_data' in the
    user home folder.

    Alternatively, it can be set programmatically by giving an explicit folder
    path. The '~' symbol is expanded to the user home folder.

    If the folder does not already exist, it is automatically created.

    Parameters
    ----------
    data_home : pathlib.Path or string, optional
        The path to the skrub data directory. If `None`, the default path
        is `~/skrub_data`.

    Returns
    -------
    data_home : pathlib.Path
        The validated path to the skrub data directory.
    """
    if data_home is None:
        data_home = Path("~").expanduser() / "skrub_data"
    else:
        data_home = Path(data_home)
    data_home = data_home.resolve()
    data_home.mkdir(parents=True, exist_ok=True)
    return data_home


def get_data_dir(name=None, data_home=None):
    """
    Returns the directory in which skrub looks for data.

    This is typically useful for the end-user to check
    where the data is downloaded and stored.

    Parameters
    ----------
    name : str, optional
        Subdirectory name. If omitted, the root data directory is returned.
    data_home : pathlib.Path or str, optional
        The path to skrub data directory. If `None`, the default path
        is `~/skrub_data`.
    """
    data_dir = get_data_home(data_home)
    if name is not None:
        data_dir = data_dir / name
    return data_dir



import re
from tempfile import TemporaryDirectory
from unittest import mock
from urllib.error import URLError

import pandas as pd
import pytest

from skrub.datasets import _fetching


def _has_data_id(call, data_id):
    # Unpacking copied from `mock._Call.__eq__`
    if len(call) == 2:
        args, kwargs = call
    else:
        name, args, kwargs = call
    return kwargs["data_id"] == data_id


@mock.patch(
    "skrub.datasets._fetching.fetch_openml",
    side_effect=_fetching.fetch_openml,
)
def test_openml_fetching(fetch_openml_mock):
    """
    Downloads a small dataset (midwest survey) and performs a bunch of tests
    that asserts the fetching function works correctly.

    We don't download all datasets as it would take way too long:
    see https://github.com/skrub-data/skrub/issues/523
    """
    with TemporaryDirectory() as temp_dir:
        # Download the dataset without loading it in memory.
        dataset = _fetching.fetch_midwest_survey(
            data_directory=temp_dir,
            load_dataframe=False,
        )
        fetch_openml_mock.assert_called_once()
        assert _has_data_id(fetch_openml_mock.call_args, _fetching.MIDWEST_SURVEY_ID)
        fetch_openml_mock.reset_mock()
        assert isinstance(dataset, _fetching.DatasetInfoOnly)
        # Briefly load the dataframe in memory, to test that reading
        # from disk works.
        assert pd.read_csv(dataset.path, **dataset.read_csv_kwargs).shape == (2494, 29)
        assert dataset.name == (
            _fetching.fetch_midwest_survey.__name__[len("fetch_") :]
            .replace("_", " ")
            .capitalize()
        )
        assert dataset.source.startswith("https://www.openml.org/")
        assert str(_fetching.MIDWEST_SURVEY_ID) in dataset.source
        # Now, load it into memory, and expect `fetch_openml`
        # to not be called because the dataset is already on disk.
        dataset = _fetching.fetch_midwest_survey(data_directory=temp_dir)
        fetch_openml_mock.assert_not_called()
        fetch_openml_mock.reset_mock()
        assert dataset.X.shape == (2494, 28)
        assert dataset.y.shape == (2494,)


def test_openml_datasets_exist():
    """
    Queries OpenML to see if the datasets are still available on the website.
    """
    openml = pytest.importorskip("openml")
    openml.datasets.check_datasets_active(
        dataset_ids=[
            _fetching.ROAD_SAFETY_ID,
            _fetching.OPEN_PAYMENTS_ID,
            _fetching.MIDWEST_SURVEY_ID,
            _fetching.MEDICAL_CHARGE_ID,
            _fetching.EMPLOYEE_SALARIES_ID,
            _fetching.TRAFFIC_VIOLATIONS_ID,
            _fetching.DRUG_DIRECTORY_ID,
        ],
        raise_error_if_not_exist=True,
    )


@mock.patch("skrub.datasets._fetching.fetch_openml")
def test_openml_datasets_calls(fetch_openml_mock):
    """
    Checks that calling the fetching functions actually calls
    `sklearn.datasets.fetch_openml`.
    Complementary to `test_openml_fetching`
    """
    with TemporaryDirectory() as temp_dir:
        for fetching_function, identifier in [
            (_fetching.fetch_road_safety, _fetching.ROAD_SAFETY_ID),
            (_fetching.fetch_open_payments, _fetching.OPEN_PAYMENTS_ID),
            (_fetching.fetch_midwest_survey, _fetching.MIDWEST_SURVEY_ID),
            (_fetching.fetch_medical_charge, _fetching.MEDICAL_CHARGE_ID),
            (_fetching.fetch_employee_salaries, _fetching.EMPLOYEE_SALARIES_ID),
            (_fetching.fetch_traffic_violations, _fetching.TRAFFIC_VIOLATIONS_ID),
            (_fetching.fetch_drug_directory, _fetching.DRUG_DIRECTORY_ID),
        ]:
            try:
                fetching_function(data_directory=temp_dir)
            except FileNotFoundError:
                pass
            fetch_openml_mock.assert_called_once()
            assert _has_data_id(fetch_openml_mock.call_args, identifier)
            fetch_openml_mock.reset_mock()


def test_fetch_world_bank_indicator():
    """
    Tests the ``fetch_world_bank_indicator()``
    function in a real environment.
    """
    test_dataset = {
        "id": "NY.GDP.PCAP.CD",
        "desc_start": "This table shows",
        "url": "https://api.worldbank.org/v2/en/indicator/NY.GDP.PCAP.CD?downloadformat=csv",  # noqa
        "dataset_columns_count": 2,
    }

    with TemporaryDirectory() as temp_dir:
        try:
            # First, we want to purposefully test FileNotFoundError exceptions.
            with pytest.raises(FileNotFoundError):
                assert _fetching.fetch_world_bank_indicator(
                    indicator_id=0, data_directory=temp_dir
                )
                assert _fetching.fetch_world_bank_indicator(
                    indicator_id=2**32,
                    data_directory=temp_dir,
                )

            # Valid call
            returned_info = _fetching.fetch_world_bank_indicator(
                indicator_id=test_dataset["id"],
                data_directory=temp_dir,
            )

        except (ConnectionError, URLError):
            pytest.skip(
                "Exception: Skipping this test because we encountered an "
                "issue probably related to an Internet connection problem. "
            )
            return

        assert returned_info.description.startswith(test_dataset["desc_start"])
        assert returned_info.source == test_dataset["url"]
        assert returned_info.path.is_file()

        dataset = pd.read_csv(returned_info.path)

        assert dataset.columns[0] == "Country Name"
        assert dataset.shape[1] == test_dataset["dataset_columns_count"]

        # Now that we have verified the file is on disk, we want to test
        # whether calling the function again reads it from disk (it should)
        # or queries the network again (it shouldn't).
        with mock.patch("urllib.request.urlretrieve") as mock_urlretrieve:
            # Same valid call as above
            disk_loaded_info = _fetching.fetch_world_bank_indicator(
                indicator_id=test_dataset["id"],
                data_directory=temp_dir,
            )
            mock_urlretrieve.assert_not_called()
            assert disk_loaded_info == returned_info


@pytest.mark.xfail(run=False, reason="temporary SSL certificate issue")
def test_fetch_movielens():
    """
    Tests the ``fetch_movielens()`` function in a real environment.
    """
    test_dataset = {
        "id": "ratings",
        "desc_start": "Summary\n=======\n\nThis dataset (ml-latest-small)",
        "url": (
            "https://files.grouplens.org/datasets/movielens/ml-latest-small.zip"
        ),  # noqa
        "dataset_columns_count": 4,
    }

    with TemporaryDirectory() as temp_dir:
        try:
            # First, we want to purposefully test ValueError exceptions.
            msg = "dataset_id options are ['movies', 'ratings'], got 'wrong_name'."
            with pytest.raises(ValueError, match=re.escape(msg)):
                assert _fetching.fetch_movielens(
                    dataset_id="wrong_name", data_directory=temp_dir
                )

            # Valid call
            returned_info = _fetching.fetch_movielens(
                dataset_id="ratings",
                data_directory=temp_dir,
            )

        except (ConnectionError, URLError):
            pytest.skip(
                "Exception: Skipping this test because we encountered an "
                "issue probably related to an Internet connection problem. "
            )
            return

        assert returned_info.description.startswith(test_dataset["desc_start"])
        assert returned_info.source == test_dataset["url"]
        assert returned_info.path.is_file()

        dataset = pd.read_csv(returned_info.path)

        assert dataset.columns[0] == "userId"
        assert dataset.shape[1] == test_dataset["dataset_columns_count"]

        # Now that we have verified the file is on disk, we want to test
        # whether calling the function again reads it from disk (it should)
        # or queries the network again (it shouldn't).
        with mock.patch("urllib.request.urlretrieve") as mock_urlretrieve:
            # Same valid call as above
            disk_loaded_info = _fetching.fetch_movielens(
                dataset_id=test_dataset["id"],
                data_directory=temp_dir,
            )
            mock_urlretrieve.assert_not_called()
            assert disk_loaded_info == returned_info



import shutil
import tempfile
from pathlib import Path

import pytest

from skrub.datasets._utils import get_data_dir, get_data_home


@pytest.mark.parametrize("data_home_type", ["string", "path"])
def test_get_data_dir(data_home_type):
    with tempfile.TemporaryDirectory() as target_dir:
        tmp_dir = target_dir if data_home_type == "string" else Path(target_dir)

        # with a pre-existing folder
        assert Path(target_dir).exists()
        data_home = get_data_home(data_home=tmp_dir)
        assert data_home == Path(target_dir).resolve()

        assert data_home.exists()

        assert (
            get_data_dir(name="tests", data_home=tmp_dir)
            == data_home.resolve() / "tests"
        )

        # if the folder is missing it will be created
        shutil.rmtree(tmp_dir)
        assert not Path(target_dir).exists()
        data_home = get_data_home(data_home=tmp_dir)
        assert data_home.exists()


def test_get_data_home_default():
    """Test function for ``get_data_home()`` with default `data_home`."""
    # We should take care of not deleting the folder if our user
    # already cached some data
    user_path = Path("~").expanduser() / "skrub_data"
    is_already_existing = user_path.exists()

    data_home = get_data_home(data_home=None)
    assert data_home == user_path
    assert data_home.exists()

    if not is_already_existing:
        # Clear the folder if it was not already existing.
        shutil.rmtree(user_path)
        assert not user_path.exists()



import pandas as pd
import pytest

from skrub import _dataframe as sbd
from skrub._agg_joiner import AggJoiner, AggTarget, split_num_categ_operations
from skrub.conftest import _POLARS_INSTALLED


@pytest.fixture
def main_table():
    df = pd.DataFrame(
        {
            "userId": [1, 1, 1, 2, 2, 2],
            "movieId": [1, 3, 6, 318, 6, 1704],
            "rating": [4.0, 4.0, 4.0, 3.0, 2.0, 4.0],
            "genre": ["drama", "drama", "comedy", "sf", "comedy", "sf"],
        }
    )
    return df


def test_split_num_categ_operations():
    "Check that the operations are correctly separated."

    # Check split ops
    num_ops, categ_ops = split_num_categ_operations(
        ["mean", "std", "min", "mode", "max", "sum"]
    )
    assert num_ops == ["mean", "std", "min", "max", "sum"]
    assert categ_ops == ["mode"]


@pytest.mark.parametrize("use_X_placeholder", [False, True])
def test_simple_fit_transform(df_module, main_table, use_X_placeholder):
    "Check the general behaviour of the `AggJoiner`."
    X = df_module.DataFrame(main_table)
    aux = X if not use_X_placeholder else "X"

    agg_joiner_user = AggJoiner(
        aux_table=aux,
        aux_key="userId",
        main_key="userId",
        cols=["rating", "genre"],
        suffix="_user",
    )

    main_user = agg_joiner_user.fit_transform(X)

    expected_user = df_module.make_dataframe(
        {
            "userId": [1, 1, 1, 2, 2, 2],
            "movieId": [1, 3, 6, 318, 6, 1704],
            "rating": [4.0, 4.0, 4.0, 3.0, 2.0, 4.0],
            "genre": ["drama", "drama", "comedy", "sf", "comedy", "sf"],
            "genre_mode_user": ["drama", "drama", "drama", "sf", "sf", "sf"],
            "rating_mean_user": [4.0, 4.0, 4.0, 3.0, 3.0, 3.0],
        }
    )
    df_module.assert_frame_equal(
        sbd.pandas_convert_dtypes(main_user), sbd.pandas_convert_dtypes(expected_user)
    )

    agg_joiner_movie = AggJoiner(
        aux_table=aux,
        aux_key="movieId",
        main_key="movieId",
        cols=["rating"],
        suffix="_movie",
    )

    main_movie = agg_joiner_movie.fit_transform(X)

    expected_movie = df_module.make_dataframe(
        {
            "userId": [1, 1, 1, 2, 2, 2],
            "movieId": [1, 3, 6, 318, 6, 1704],
            "rating": [4.0, 4.0, 4.0, 3.0, 2.0, 4.0],
            "genre": ["drama", "drama", "comedy", "sf", "comedy", "sf"],
            "rating_mean_movie": [4.0, 4.0, 3.0, 3.0, 3.0, 4.0],
        }
    )

    df_module.assert_frame_equal(
        sbd.pandas_convert_dtypes(main_movie), sbd.pandas_convert_dtypes(expected_movie)
    )


def test_correct_keys(df_module, main_table):
    "Check that expected `key` parameters for the `AggJoiner` are working."
    main_table = df_module.DataFrame(main_table)

    # Check only key
    agg_joiner = AggJoiner(
        aux_table=main_table,
        key="userId",
        cols=["rating", "genre"],
    )
    agg_joiner.fit(main_table)
    assert agg_joiner._main_key == ["userId"]
    assert agg_joiner._aux_key == ["userId"]

    # Check multiple key
    agg_joiner = AggJoiner(
        aux_table=main_table,
        key=["userId", "movieId"],
        cols=["rating", "genre"],
    )
    agg_joiner.fit(main_table)

    # Check multiple main_key and aux_key, same length
    agg_joiner = AggJoiner(
        aux_table=main_table,
        main_key=["userId", "movieId"],
        aux_key=["userId", "movieId"],
        cols=["rating", "genre"],
    )
    agg_joiner.fit(main_table)
    assert agg_joiner._main_key == ["userId", "movieId"]
    assert agg_joiner._aux_key == ["userId", "movieId"]


def test_wrong_keys(df_module, main_table):
    "Check that wrong `key` parameters for the `AggJoiner` raise an error."
    main_table = df_module.DataFrame(main_table)

    # Check too many main_key
    agg_joiner = AggJoiner(
        aux_table=main_table,
        main_key=["userId", "movieId"],
        aux_key="userId",
        cols=["rating", "genre"],
    )
    with pytest.raises(
        ValueError, match=r"(?=.*Cannot join on different numbers of columns)"
    ):
        agg_joiner.fit(main_table)

    # Check too many aux_key
    agg_joiner = AggJoiner(
        aux_table=main_table,
        main_key="userId",
        aux_key=["userId", "movieId"],
        cols=["rating", "genre"],
    )
    with pytest.raises(
        ValueError, match=r"(?=.*Cannot join on different numbers of columns)"
    ):
        agg_joiner.fit(main_table)

    # Check providing key and extra main_key
    agg_joiner = AggJoiner(
        aux_table=main_table,
        key="userId",
        main_key="userId",
        cols=["rating", "genre"],
    )
    with pytest.raises(ValueError, match=r"(?=.*not a combination of both.)"):
        agg_joiner.fit(main_table)

    # Check providing key and extra aux_key
    agg_joiner = AggJoiner(
        aux_table=main_table,
        key="userId",
        aux_key="userId",
        cols=["rating", "genre"],
    )
    with pytest.raises(ValueError, match=r"(?=.*not a combination of both.)"):
        agg_joiner.fit(main_table)

    # Check main_key doesn't exist in table
    agg_joiner = AggJoiner(
        aux_table=main_table,
        main_key="wrong_key",
        aux_key="userId",
        cols=["rating", "genre"],
    )
    match = r"(?=.*columns cannot be used because they do not exist)"
    with pytest.raises(ValueError, match=match):
        agg_joiner.fit(main_table)

    # Check aux_key doesn't exist in table
    agg_joiner = AggJoiner(
        aux_table=main_table,
        main_key="userId",
        aux_key="wrong_key",
        cols=["rating", "genre"],
    )
    match = r"(?=.*columns cannot be used because they do not exist)"
    with pytest.raises(ValueError, match=match):
        agg_joiner.fit(main_table)


def test_default_suffix(df_module, main_table):
    "Check that the default `suffix` of `AggJoiner` is ''."
    main_table = df_module.DataFrame(main_table)

    # Check no suffix
    agg_joiner = AggJoiner(
        aux_table=main_table,
        key="userId",
        cols=["rating", "genre"],
    )
    agg_joiner.fit(main_table)
    assert agg_joiner.suffix == ""


def test_too_many_suffixes(df_module, main_table):
    "Check that providing more than one `suffix` for the `AggJoiner` raises an error."
    main_table = df_module.DataFrame(main_table)

    # Check inconsistent number of suffixes
    agg_joiner = AggJoiner(
        aux_table=main_table,
        key="userId",
        cols="rating",
        suffix=["_user", "_movie", "_tag"],
    )
    with pytest.raises(ValueError, match=r"(?='suffix' must be a string.*)"):
        agg_joiner.fit(main_table)


def test_default_cols(df_module, main_table):
    "Check that by default, `cols` are all the columns of `aux_table` except `aux_key`."
    main_table = df_module.DataFrame(main_table)

    # Check no cols
    agg_joiner = AggJoiner(
        aux_table=main_table,
        key=["movieId", "userId"],
    )
    agg_joiner.fit(main_table)
    agg_joiner._cols == ["rating", "genre"]


def test_correct_cols(df_module, main_table):
    "Check that expected `cols` parameters for the `AggJoiner` are working."
    main_table = df_module.DataFrame(main_table)

    # Check one col
    agg_joiner = AggJoiner(
        aux_table=main_table,
        key=["movieId", "userId"],
        cols=["rating"],
    )
    agg_joiner.fit(main_table)
    agg_joiner._cols == ["rating"]


def test_wrong_cols(df_module, main_table):
    "Check that providing a column that's not in `aux_table` does not work."
    main_table = df_module.DataFrame(main_table)

    # Check missing agg or keys cols in tables
    agg_joiner = AggJoiner(
        aux_table=main_table,
        key="userId",
        cols="unknown_col",
    )
    match = r"(?=.*columns cannot be used because they do not exist)"
    with pytest.raises(ValueError, match=match):
        agg_joiner.fit(main_table)


def test_input_multiple_tables(df_module, main_table):
    "Check that providing too many auxiliary tables in `AggJoiner` raises an error."
    main_table = df_module.DataFrame(main_table)

    # Check too many aux_table
    agg_joiner = AggJoiner(
        aux_table=[main_table, main_table],
        main_key="userId",
        aux_key=["userId", "userId"],
        cols=[["rating"], ["rating"]],
    )
    error_msg = (
        r"(?=.*must be a dataframe or the string 'X')"
        r"(?=.*use the MultiAggJoiner instead)"
    )
    with pytest.raises(TypeError, match=error_msg):
        agg_joiner.fit_transform(main_table)


def test_default_operations(df_module, main_table):
    "Check that the default `operations` of `AggJoiner` are ['mean', 'mode']."
    main_table = df_module.DataFrame(main_table)

    # Check default operations
    agg_joiner = AggJoiner(
        aux_table=main_table,
        key="userId",
        cols=["rating", "genre"],
    )
    agg_joiner.fit(main_table)
    assert agg_joiner._operations == ["mean", "mode"]


def test_correct_operations_input(df_module, main_table):
    "Check that expected `operations` parameters for the `AggJoiner` are working."
    main_table = df_module.DataFrame(main_table)

    # Check invariant operations input
    agg_joiner = AggJoiner(
        aux_table=main_table,
        key="userId",
        cols=["rating", "genre"],
        operations=["min", "max", "mode"],
    )
    agg_joiner.fit(main_table)
    assert agg_joiner._operations == ["min", "max", "mode"]


@pytest.mark.skipif(not _POLARS_INSTALLED, reason="Polars is not available")
def test_polars_unavailable_operation(main_table):
    "Check that the 'value_counts' operation is not supported for Polars."
    import polars as pl

    agg_joiner = AggJoiner(
        aux_table="X",
        main_key="userId",
        aux_key="movieId",
        cols="rating",
        operations=["value_counts"],
    )
    with pytest.raises(ValueError, match=r"(?=.*value_counts)(?=.*supported)"):
        agg_joiner.fit(pl.DataFrame(main_table))


def test_not_supported_operations(df_module, main_table):
    "Check that calling an unsupported operation raises an error."
    main_table = df_module.DataFrame(main_table)

    # Check not supported operations
    agg_joiner = AggJoiner(
        aux_table=main_table,
        key="userId",
        cols=["rating", "genre"],
        operations=["nunique", "mode"],
    )
    match = r"(?=.*operations options are)"
    with pytest.raises(ValueError, match=match):
        agg_joiner.fit(main_table)


def test_wrong_string_placeholder(df_module, main_table):
    "Check that `aux_table='Y'` is not a valid string placeholder."
    main_table = df_module.DataFrame(main_table)

    agg_joiner = AggJoiner(
        aux_table="Y",
        key="userId",
        cols="genre",
    )
    with pytest.raises(ValueError, match=r"(?=.*dataframe)(?=.*'X')"):
        agg_joiner.fit(main_table)


def test_not_fitted_dataframe(df_module, main_table):
    """
    Check that calling `transform` on a dataframe not containing the columns
    seen during `fit` raises an error.
    """
    main_table = df_module.DataFrame(main_table)
    not_main_table = df_module.make_dataframe(
        {"wrong": [1, 2, 3], "dataframe": [4, 5, 6]}
    )

    agg_joiner = AggJoiner(
        aux_table=main_table,
        key="userId",
    )
    agg_joiner.fit(main_table)
    error_msg = r"(?=.*columns cannot be used because they do not exist)"
    with pytest.raises(ValueError, match=error_msg):
        agg_joiner.transform(not_main_table)


y = pd.DataFrame(dict(rating=[4.0, 4.0, 4.0, 3.0, 2.0, 4.0]))


@pytest.mark.parametrize(
    "y, col_name",
    [
        (y, "rating"),
        (y["rating"], "rating"),
        (y.values, "y_0"),
        (y.values.tolist(), "y_0"),
    ],
)
def test_agg_target(main_table, y, col_name):
    agg_target = AggTarget(
        main_key="userId",
        suffix="_user",
        operation=["hist(2)", "value_counts"],
    )
    main_transformed = agg_target.fit_transform(main_table, y)

    main_transformed_expected = pd.DataFrame(
        {
            "userId": [1, 1, 1, 2, 2, 2],
            "movieId": [1, 3, 6, 318, 6, 1704],
            "rating": [4.0, 4.0, 4.0, 3.0, 2.0, 4.0],
            "genre": ["drama", "drama", "comedy", "sf", "comedy", "sf"],
            "index": [0, 0, 0, 1, 1, 1],
            f"{col_name}_(1.999, 3.0]_user": [0, 0, 0, 2, 2, 2],
            f"{col_name}_(3.0, 4.0]_user": [3, 3, 3, 1, 1, 1],
            f"{col_name}_2.0_user": [0.0, 0.0, 0.0, 1.0, 1.0, 1.0],
            f"{col_name}_3.0_user": [0.0, 0.0, 0.0, 1.0, 1.0, 1.0],
            f"{col_name}_4.0_user": [3.0, 3.0, 3.0, 1.0, 1.0, 1.0],
        }
    )
    pd.testing.assert_frame_equal(main_transformed, main_transformed_expected)


def test_agg_target_missing_operations(main_table):
    agg_target = AggTarget(
        main_key="userId",
        suffix="_user",
    )

    # y is continuous
    y = pd.DataFrame(dict(rating=[4.0, 4.1, 4.2, 4.3, 4.4, 4.5]))
    agg_target.fit(main_table, y)
    assert agg_target.operation_ == ["mean"]

    # y is categorical
    y = pd.DataFrame(dict(rating=["1", "2", "3", "1", "2", "3"]))
    agg_target.fit(main_table, y)
    assert agg_target.operation_ == ["mode"]


def test_agg_target_check_input(main_table):
    agg_target = AggTarget(
        main_key="userId",
        suffix="_user",
    )
    match = r"(?=.*X must be a dataframe)"
    with pytest.raises(TypeError, match=match):
        agg_target.fit(main_table.values, y)

    match = r"(?=.*length)(?=.*match)"
    with pytest.raises(ValueError, match=match):
        agg_target.fit(main_table, y["rating"][:2])


def test_no_aggregation_exception(main_table):
    agg_target = AggTarget(
        main_key="userId",
        operation=[],
    )
    with pytest.raises(ValueError, match=r"(?=.*No aggregation)"):
        agg_target.fit(main_table, y)


def test_wrong_args_ops(main_table):
    agg_target = AggTarget(
        main_key="userId",
        operation="mean(2)",
    )
    with pytest.raises(ValueError, match=r"(?=.*'mean')(?=.*argument)"):
        agg_target.fit(main_table, y)



"""
Self-aggregation on MovieLens
=============================

MovieLens is a famous movie dataset used for both explicit
and implicit recommender systems. It provides a main table,
"ratings", that can be viewed as logs or transactions, comprised
of only 4 columns: ``userId``, ``movieId``, ``rating`` and ``timestamp``.
MovieLens also gives a contextual table "movies", including
``movieId``, ``title`` and ``types``, to enable content-based feature extraction.

From the perspective of machine-learning pipelines, one challenge is to
transform the transaction log into features that can be fed to supervised learning.

In this notebook, we only deal with the main table "ratings".
Our objective is **not to achieve state-of-the-art performance** on
the explicit regression task, but rather to illustrate how to perform
feature engineering in a simple way using |AggJoiner| and |AggTarget|.
Note that our performance is higher than the baseline of using the mean
rating per movies.

The benefit of using  |AggJoiner| and |AggTarget| is that they readily
provide a full pipeline, from the original tables to the prediction, that can
be cross-validated or applied to new data to serve prediction. At the end of
this example, we showcase hyper-parameter optimization on the whole pipeline.


.. |AggJoiner| replace::
     :class:`~skrub.AggJoiner`

.. |AggTarget| replace::
     :class:`~skrub.AggTarget`

.. |TableVectorizer| replace::
     :class:`~skrub.TableVectorizer`

.. |DatetimeEncoder| replace::
     :class:`~skrub.DatetimeEncoder`

.. |TargetEncoder| replace::
     :class:`~sklearn.preprocessing.TargetEncoder`

.. |make_pipeline| replace::
     :class:`~sklearn.pipeline.make_pipeline`

.. |Pipeline| replace::
     :class:`~sklearn.pipeline.Pipeline`

.. |GridSearchCV| replace::
     :class:`~sklearn.model_selection.GridSearchCV`

.. |TimeSeriesSplit| replace::
     :class:`~sklearn.model_selection.TimeSeriesSplit`

.. |HGBR| replace::
     :class:`~sklearn.ensemble.HistGradientBoostingRegressor`
"""

###############################################################################
# The data
# --------
#
# We begin with loading the ratings table from MovieLens.
# Note that we use the light version (100k rows).
import pandas as pd

from skrub.datasets import fetch_movielens

ratings = fetch_movielens(dataset_id="ratings")
ratings = ratings.X.sort_values("timestamp").reset_index(drop=True)
ratings["timestamp"] = pd.to_datetime(ratings["timestamp"], unit="s")

X = ratings[["userId", "movieId", "timestamp"]]
y = ratings["rating"]
X.shape, y.shape
###############################################################################
X.head()

###############################################################################
# Encoding the timestamp with a TableVectorizer
# ---------------------------------------------
#
# Our first step is to extract features from the timestamp, using the
# |TableVectorizer|. Natively, it uses the |DatetimeEncoder| on datetime
# columns, and doesn't interact with numerical columns.
from skrub import DatetimeEncoder, TableVectorizer

table_vectorizer = TableVectorizer(datetime=DatetimeEncoder(add_weekday=True))
X_date_encoded = table_vectorizer.fit_transform(X)
X_date_encoded.head()

###############################################################################
# We can now make a couple of plots and gain some insight on our dataset.
import seaborn as sns
from matplotlib import pyplot as plt

sns.set_style("darkgrid")


def make_barplot(x, y, title):
    fig, ax = plt.subplots(layout="constrained")
    norm = plt.Normalize(y.min(), y.max())
    cmap = plt.get_cmap("magma")

    sns.barplot(x=x, y=y, palette=cmap(norm(y)), ax=ax)
    ax.set_title(title)
    ax.set_xticks(ax.get_xticks(), labels=ax.get_xticklabels(), rotation=30)
    ax.set_ylabel(None)


# O is Monday, 6 is Sunday

daily_volume = X_date_encoded["timestamp_weekday"].value_counts().sort_index()

make_barplot(
    x=daily_volume.index,
    y=daily_volume.values,
    title="Daily volume of ratings",
)

###############################################################################
# We also display the distribution of our target ``y``.
rating_count = y.value_counts().sort_index()

make_barplot(
    x=rating_count.index,
    y=rating_count.values,
    title="Distribution of ratings given to movies",
)


###############################################################################
# AggTarget: aggregate y, then join
# ---------------------------------
#
# We have just extracted datetime features from timestamps.
#
# Let's now perform an expansion for the target ``y``, by aggregating it before
# joining it back on the main table. The biggest risk of doing target expansion
# with multiple dataframe operations yourself is to end up leaking the target.
#
# To solve this, the |AggTarget| transformer allows you to
# aggregate the target ``y`` before joining it on the main table, without
# risk of leaking. Note that to perform aggregation then joining on the features
# ``X``, you need to use |AggJoiner| instead.
#
# You can also think of it as a generalization of the |TargetEncoder|, which
# encodes categorical features based on the target.
#
# We only focus on aggregating the target by **users**, but later we will
# also consider aggregating by **movies**. Here, we compute the histogram of the
# target with 3 bins, before joining it back on the initial table.
#
# This feature answer questions like
# *"How many times has this user given a bad, medium or good rate to movies?"*.
from skrub import AggTarget

agg_target_user = AggTarget(
    main_key="userId",
    suffix="_user",
    operation="hist(3)",
)
X_transformed = agg_target_user.fit_transform(X, y)

X_transformed.shape
###############################################################################
X_transformed.head()

###############################################################################
# Similarly, we join on ``movieId`` instead of ``userId``.
#
# This feature answer questions like
# *"How many times has this movie received a bad, medium or good rate from users?"*.
agg_target_movie = AggTarget(
    main_key="movieId",
    suffix="_movie",
    operation="hist(3)",
)
X_transformed = agg_target_movie.fit_transform(X, y)
X_transformed.shape
###############################################################################
X_transformed.head()

###############################################################################
# Chaining everything together in a pipeline
# ------------------------------------------
#
# To perform cross-validation and enable hyper-parameter tuning, we gather
# all elements into a scikit-learn |Pipeline| by using |make_pipeline|,
# and define a scikit-learn |HGBR|.
from sklearn.ensemble import HistGradientBoostingRegressor
from sklearn.pipeline import make_pipeline

pipeline = make_pipeline(
    table_vectorizer,
    agg_target_user,
    agg_target_movie,
    HistGradientBoostingRegressor(learning_rate=0.1, max_depth=4, max_iter=40),
)

pipeline

###############################################################################
# Hyper-parameters tuning and cross validation
# --------------------------------------------
#
# We can finally create our hyper-parameter search space, and use a
# |GridSearchCV|. We select the cross validation splitter to be
# the |TimeSeriesSplit| to prevent leakage, since our data are timestamped
# logs.
#
# Note that you need the name of the pipeline elements to assign them
# hyper-parameters search.
#
# You can lookup the name of the pipeline elements by doing:
list(pipeline.named_steps)

###############################################################################
# Alternatively, you can use scikit-learn |Pipeline| to name your transformers:
# ``Pipeline([("agg_target_user", agg_target_user), ...])``
#
# We now perform the grid search over the ``AggTarget`` transformers to find the
# operation maximizing our validation score.
from sklearn.model_selection import GridSearchCV, TimeSeriesSplit

operations = ["mean", "hist(3)", "hist(5)", "hist(7)", "value_counts"]
param_grid = [
    {
        "aggtarget-2__operation": [op],
    }
    for op in operations
]

cv = GridSearchCV(pipeline, param_grid, cv=TimeSeriesSplit(n_splits=10))
cv.fit(X, y)

results = pd.DataFrame(cv.cv_results_)

cols = [f"split{idx}_test_score" for idx in range(10)]
results = results.set_index("param_aggtarget-2__operation")[cols].T
results

###############################################################################
# The score used in this regression task is the R2. Remember that the R2
# evaluates the relative performance compared to the naive baseline consisting
# in always predicting the mean value of ``y_test``.
# Therefore, the R2 is 0 when ``y_pred = y_true.mean()`` and is upper bounded
# to 1 when ``y_pred = y_true``.
#
# To get a better sense of the learning performances of our simple pipeline,
# we also compute the average rating of each movie in the training set,
# and uses this average to predict the ratings in the test set.
from sklearn.metrics import r2_score


def baseline_r2(X, y, train_idx, test_idx):
    """Compute the average rating for all movies in the train set,
    and map these averages to the test set as a prediction.

    If a movie in the test set is not present in the training set,
    we simply predict the global average rating of the training set.
    """
    X_train, y_train = X.iloc[train_idx].copy(), y.iloc[train_idx]
    X_test, y_test = X.iloc[test_idx], y.iloc[test_idx]

    X_train["y"] = y_train

    movie_avg_rating = X_train.groupby("movieId")["y"].mean().to_frame().reset_index()

    y_pred = X_test.merge(movie_avg_rating, on="movieId", how="left")["y"]
    y_pred = y_pred.fillna(y_pred.mean())

    return r2_score(y_true=y_test, y_pred=y_pred)


all_baseline_r2 = []
for train_idx, test_idx in TimeSeriesSplit(n_splits=10).split(X, y):
    all_baseline_r2.append(baseline_r2(X, y, train_idx, test_idx))

results.insert(0, "naive mean estimator", all_baseline_r2)

# we only keep the 5 out of 10 last results
# because the initial size of the train set is rather small
fig, ax = plt.subplots(layout="constrained")
sns.boxplot(results.tail(5), palette="magma", ax=ax)
ax.set_ylabel("R2 score")
ax.set_title("Hyper parameters grid-search results")
plt.tight_layout()

###############################################################################
# The naive estimator has a lower performance than our pipeline, which means
# that our extracted features brought some predictive power.
#
# It seems that using the ``"value_counts"`` as an aggregation operator for
# |AggTarget| yields better performances than using the mean (which is
# equivalent to using the |TargetEncoder|).
#
# Here, the number of bins encoding the target is proportional to the
# performance: computing the mean yields a single statistic, whereas histograms
# yield a density over a reduced set of bins, and ``"value_counts"`` yields an
# exhaustive histogram over all the possible values of ratings
# (here 10 different values, from 0.5 to 5).



"""
Pandas specialization of the aggregate and join operations.
"""
import re
from itertools import product

import numpy as np
import pandas as pd

from skrub._utils import atleast_1d_or_none

# TODO: _dataframe._pandas is temporary; all code in this module should be moved
# elsewhere and use the dispatch mechanism.


def aggregate(
    table,
    key,
    cols_to_agg,
    num_operations=("mean",),
    categ_operations=("mode",),
    suffix=None,
):
    """Aggregates a :obj:`pandas.DataFrame`.

    This function uses the ``dataframe.groupby(key).agg`` method from Pandas.

    Parameters
    ----------
    table : pd.DataFrame
        The input dataframe to aggregate.

    key : str or Iterable[str]
        The columns used as keys to aggregate on.

    cols_to_agg : str or Iterable[str]
        The columns to aggregate.

    num_operations : str or Iterable[str], default=("mean",)
        The reduction functions to apply on numerical columns
        in ``cols_to_agg`` during the aggregation.

    categ_operations : str or Iterable[str], default=("mode",)
        The reduction functions to apply on categorical columns
        in ``cols_to_agg`` during the aggregation.

    suffix : str, optional
        The suffix appended to output columns.

    Returns
    -------
    group : pd.DataFrame
        The aggregated output.
    """
    if not isinstance(table, pd.DataFrame):
        raise TypeError(f"'table' must be a pandas dataframe, got {type(table)!r}.")

    key = atleast_1d_or_none(key)
    cols_to_agg = atleast_1d_or_none(cols_to_agg)
    num_operations = atleast_1d_or_none(num_operations)
    categ_operations = atleast_1d_or_none(categ_operations)
    suffix = "" if suffix is None else suffix

    num_cols, categ_cols = split_num_categ_cols(table[cols_to_agg])

    num_named_agg, num_value_counts = get_named_agg(table, num_cols, num_operations)
    categ_named_agg, categ_value_counts = get_named_agg(
        table, categ_cols, categ_operations
    )

    named_agg = {**num_named_agg, **categ_named_agg}
    if named_agg:
        base_group = table.groupby(key).agg(**named_agg)
    else:
        base_group = None

    # 'histogram' and 'value_counts' requires a pivot
    value_counts = {**num_value_counts, **categ_value_counts}
    for output_key, (col_to_agg, kwargs) in value_counts.items():
        serie_group = table.groupby(key)[col_to_agg].value_counts(**kwargs)
        serie_group.name = output_key
        pivot = (
            serie_group.reset_index()
            .pivot(index=key, columns=col_to_agg)
            .reset_index()
            .fillna(0)
        )
        cols = pivot.columns.droplevel(0)
        index_cols = np.atleast_1d(key).tolist()
        feature_cols = (f"{col_to_agg}_" + cols[len(index_cols) :].astype(str)).tolist()
        cols = [*index_cols, *feature_cols]
        pivot.columns = cols

        if base_group is None:
            base_group = pivot
        else:
            base_group = base_group.merge(pivot, on=key, how="left")

    if base_group is None:
        raise ValueError("No aggregation to perform.")

    base_group.columns = [
        f"{col}{suffix}" if col not in key else col for col in base_group.columns
    ]
    sorted_cols = sorted(base_group.columns)

    return base_group[sorted_cols].reset_index(drop=False)


def get_named_agg(table, cols, operations):
    """Map aggregation tuples to their output key.

    The dictionary has the form: output_key = (column, aggfunc).
    This is used as input for the ``dataframe.agg`` method from Pandas.

    'value_counts' and 'hist' operations require to pivot
    the tables and treated in a separate mapping.

    Parameters
    ----------
    table : pd.DataFrame
        Input dataframe, only used to compute bins values if
        'value_counts' or 'hist' are operations.

    cols : list
        The columns to aggregate.

    operations : list
        The reduce operations to perform.

    Returns
    -------
    named_agg : dict
        Named aggregation mapping.

    value_counts : dict
        ``value_counts`` operations mapping.
    """
    named_agg, value_counts = {}, {}
    for col, operation in product(cols, operations):
        op_root, bin_args = _parse_argument(operation)
        aggfunc, bin_args = _get_aggfunc(table[col], op_root, bin_args)

        output_key = f"{col}_{op_root}"
        # 'value_counts' changes the index of the resulting frame
        # and must be treated separately.
        if aggfunc == "value_counts":
            value_counts[output_key] = (col, bin_args)
        else:
            named_agg[output_key] = (col, aggfunc)

    return named_agg, value_counts


def _parse_argument(operation):
    """Split a text input into a function name and its argument.

    Parameters
    ----------
    operation : str
        The operation to parse.

    Returns
    -------
    operation_root : str
        The name of the operation before parenthesis, if any.

    bin_args : int
        The number of bin to create for 'hist' or 'value_counts'.

    Examples
    --------
    >>> _parse_argument("hist(10)")
    ('hist', 10)
    """
    split = re.split("\\(.+\\)", operation)
    op_root = split[0]
    if len(split) > 1:
        # remove op_root
        bin_args = re.split(f"^{op_root}", operation)
        bin_args = bin_args[1]
        # remove parenthesis
        bin_args = re.sub("\\(|\\)", "", bin_args)
        bin_args = int(bin_args)
        return op_root, bin_args
    else:
        return op_root, None


PANDAS_OPS_MAPPING = {
    "mode": pd.Series.mode,
    "quantile": pd.Series.quantile,
    "hist": "value_counts",
}


def _get_aggfunc(serie, op_root, n_bins):
    """Map operation roots to their pandas agg functions.

    When args is provided for histogram or value_counts,
    we create args

    Parameters
    ----------
    serie : pd.Series
        Input series, used to compute the bins if `n_bins` is provided.

    op_root : str
        Operation root, the operation without the bin argument, if any.

    n_bins : int
        The number of bin to create when 'value_counts' or 'hist' operation are used.

    Returns
    -------
    aggfunc : str or callable
        The pandas agg functions to perform.

    bins_args : dict
        The bins to create when using 'value_counts' or 'hist'.
    """
    aggfunc = PANDAS_OPS_MAPPING.get(op_root, op_root)

    if n_bins is not None:
        # histogram and value_counts
        if aggfunc == "value_counts":
            # If bins is a number, we need to set a fix bin range,
            # otherwise bins edges will be defined dynamically for
            # each rows.
            min_, max_ = serie.min(), serie.max()
            bins = np.linspace(min_, max_, n_bins + 1)
            bins_args = dict(bins=bins)
        else:
            raise ValueError(
                f"Operator {op_root!r} doesn't take any argument, got {n_bins!r}"
            )
    else:
        bins_args = {}

    return aggfunc, bins_args


def split_num_categ_cols(table):
    """Split dataframe columns between numerical and categorical."""
    num_cols = table.select_dtypes("number").columns
    categ_cols = table.select_dtypes(["object", "string", "category"]).columns

    return num_cols, categ_cols


def rename_columns(dataframe, renaming_function):
    return dataframe.rename(
        columns={c: renaming_function(c) for c in dataframe.columns}
    )



"""
.. _example_datetime_encoder :

===================================================
Handling datetime features with the DatetimeEncoder
===================================================

In this example, we illustrate how to better integrate datetime features
in machine learning models with the |DatetimeEncoder|.

This encoder breaks down passed datetime features into relevant numerical
features, such as the month, the day of the week, the hour of the day, etc.

It is used by default in the |TableVectorizer|.


.. |DatetimeEncoder| replace::
    :class:`~skrub.DatetimeEncoder`

.. |TableVectorizer| replace::
    :class:`~skrub.TableVectorizer`

.. |OneHotEncoder| replace::
    :class:`~sklearn.preprocessing.OneHotEncoder`

.. |TimeSeriesSplit| replace::
    :class:`~sklearn.model_selection.TimeSeriesSplit`

.. |ColumnTransformer| replace::
    :class:`~sklearn.compose.ColumnTransformer`

.. |make_column_transformer| replace::
    :class:`~sklearn.compose.make_column_transformer`

.. |HGBR| replace::
    :class:`~sklearn.ensemble.HistGradientBoostingRegressor`

.. |to_datetime| replace::
    :func:`~skrub.to_datetime`
"""


###############################################################################
# A problem with relevant datetime features
# -----------------------------------------
#
# We will use a dataset of air quality measurements in different cities.
# In this setting, we want to predict the NO2 air concentration, based
# on the location, date and time of measurement.

from pprint import pprint

import pandas as pd

data = pd.read_csv(
    "https://raw.githubusercontent.com/pandas-dev/pandas"
    "/main/doc/data/air_quality_no2_long.csv"
).sort_values("date.utc")
# Extract our input data (X) and the target column (y)
y = data["value"]
X = data[["city", "date.utc"]]

X

###############################################################################
# We convert the dataframe date columns using |to_datetime|. Notice how
# we don't need to specify the columns to convert.
from skrub import to_datetime

X = to_datetime(X)
X.dtypes

###############################################################################
# Encoding the features
# .....................
#
# We will construct a |ColumnTransformer| in which we will encode
# the city names with a |OneHotEncoder|, and the date
# with a |DatetimeEncoder|.
#
# During the instantiation of the |DatetimeEncoder|, we specify that we want
# to extract the day of the week, and that we don't want to extract anything
# finer than minutes. This is because we don't want to extract seconds and
# lower units, as they are probably unimportant.

from sklearn.compose import make_column_transformer
from sklearn.preprocessing import OneHotEncoder

from skrub import DatetimeEncoder

encoder = make_column_transformer(
    (OneHotEncoder(handle_unknown="ignore"), ["city"]),
    (DatetimeEncoder(add_weekday=True, resolution="minute"), "date.utc"),
    remainder="drop",
)

X_enc = encoder.fit_transform(X)
# pprint(encoder.get_feature_names_out())

###############################################################################
# We see that the encoder is working as expected: the ``"date.utc"`` column has
# been replaced by features extracting the month, day, hour, minute, day of the
# week and total second since Epoch information.

###############################################################################
# One-liner with the |TableVectorizer|
# ....................................
#
# As mentioned earlier, the |TableVectorizer| makes use of the
# |DatetimeEncoder| by default.

from skrub import TableVectorizer

table_vec = TableVectorizer().fit(X)
pprint(table_vec.get_feature_names_out())

###############################################################################
# If we want to customize the |DatetimeEncoder| inside the |TableVectorizer|,
# we can replace its default parameter with a new, custom instance:
#
# Here, for example, we want it to extract the day of the week.

table_vec = TableVectorizer(datetime=DatetimeEncoder(add_weekday=True)).fit(X)
pprint(table_vec.get_feature_names_out())

###############################################################################
# .. note:
#     For more information on how to customize the |TableVectorizer|, see
#     :ref:`sphx_glr_auto_examples_01_dirty_categories.py`.
#
# Inspecting the |TableVectorizer| further, we can check that the
# |DatetimeEncoder| is used on the correct column(s).
pprint(table_vec.transformers_)

###############################################################################
# Prediction with datetime features
# ---------------------------------
#
# For prediction tasks, we recommend using the |TableVectorizer| inside a
# pipeline, combined with a model that can use the features extracted by the
# |DatetimeEncoder|.
# Here's we'll use a |HGBR| as our learner.
#
# .. note:
#    You might need to require the experimental feature for scikit-learn
#    versions earlier than 1.0 with:
#    ```py
#    from sklearn.experimental import enable_hist_gradient_boosting
#    ```
from sklearn.ensemble import HistGradientBoostingRegressor
from sklearn.pipeline import make_pipeline

pipeline = make_pipeline(table_vec, HistGradientBoostingRegressor())

###############################################################################
# Evaluating the model
# ....................
#
# When using date and time features, we often care about predicting the future.
# In this case, we have to be careful when evaluating our model, because
# the standard settings of the cross-validation do not respect time ordering.
#
# Instead, we can use the |TimeSeriesSplit|,
# which ensures that the test set is always in the future.
from sklearn.model_selection import TimeSeriesSplit, cross_val_score

cross_val_score(
    pipeline,
    X,
    y,
    scoring="neg_mean_squared_error",
    cv=TimeSeriesSplit(n_splits=5),
)

###############################################################################
# Plotting the prediction
# .......................
#
# The mean squared error is not obvious to interpret, so we compare
# visually the prediction of our model with the actual values.
import matplotlib.pyplot as plt

mask_train = X["date.utc"] < "2019-06-01"
X_train, X_test = X.loc[mask_train], X.loc[~mask_train]
y_train, y_test = y.loc[mask_train], y.loc[~mask_train]

pipeline.fit(X_train, y_train)
y_pred = pipeline.predict(X_test)

all_cities = X_test["city"].unique()

fig, axes = plt.subplots(nrows=len(all_cities), ncols=1, figsize=(12, 9))
for ax, city in zip(axes, all_cities):
    mask_prediction = X_test["city"] == city
    date_prediction = X_test.loc[mask_prediction]["date.utc"]
    y_prediction = y_pred[mask_prediction]

    mask_reference = X["city"] == city
    date_reference = X.loc[mask_reference]["date.utc"]
    y_reference = y[mask_reference]

    ax.plot(date_reference, y_reference, label="Actual")
    ax.plot(date_prediction, y_prediction, label="Predicted")

    ax.set(
        ylabel="NO2",
        title=city,
    )
    ax.legend()

fig.subplots_adjust(hspace=0.5)
plt.show()

###############################################################################
# Let's zoom on a few days:

mask_zoom_reference = (X["date.utc"] >= "2019-06-01") & (X["date.utc"] < "2019-06-04")
mask_zoom_prediction = (X_test["date.utc"] >= "2019-06-01") & (
    X_test["date.utc"] < "2019-06-04"
)

all_cities = ["Paris", "London"]
fig, axes = plt.subplots(nrows=len(all_cities), ncols=1, figsize=(12, 9))
for ax, city in zip(axes, all_cities):
    mask_prediction = (X_test["city"] == city) & mask_zoom_prediction
    date_prediction = X_test.loc[mask_prediction]["date.utc"]
    y_prediction = y_pred[mask_prediction]

    mask_reference = (X["city"] == city) & mask_zoom_reference
    date_reference = X.loc[mask_reference]["date.utc"]
    y_reference = y[mask_reference]

    ax.plot(date_reference, y_reference, label="Actual")
    ax.plot(date_prediction, y_prediction, label="Predicted")

    ax.set(
        ylabel="NO2",
        title=city,
    )
    ax.legend()

plt.show()


###############################################################################
# Features importance
# -------------------
#
# Using the |DatetimeEncoder| allows us to better understand how the date
# impacts the NO2 concentration. To this aim, we can compute the
# importance of the features created by the |DatetimeEncoder|, using the
# :func:`~sklearn.inspection.permutation_importance` function, which
# basically shuffles a feature and sees how the model changes its prediction.

###############################################################################
from sklearn.inspection import permutation_importance

table_vec = TableVectorizer(datetime=DatetimeEncoder(add_weekday=True))

# In this case, we don't use a pipeline, because we want to compute the
# importance of the features created by the DatetimeEncoder
X_transform = table_vec.fit_transform(X)
feature_names = table_vec.get_feature_names_out()

model = HistGradientBoostingRegressor().fit(X_transform, y)
result = permutation_importance(model, X_transform, y, n_repeats=10, random_state=0)

result = pd.DataFrame(
    dict(
        feature_names=feature_names,
        std=result.importances_std,
        importances=result.importances_mean,
    )
).sort_values("importances", ascending=False)

result.plot.barh(
    y="importances", x="feature_names", title="Feature Importances", figsize=(12, 9)
)
plt.tight_layout()
plt.show()

# %%
# We can see that the total seconds since Epoch and the hour of the day
# are the most important feature, which seems reasonable.
#
# Conclusion
# ----------
#
# In this example, we saw how to use the |DatetimeEncoder| to create
# features from a date column.
# Also check out the |TableVectorizer|, which automatically recognizes
# and transforms datetime columns by default.
